from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class EffectStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("effect_type", DataType("s32"), save_as="effect_type"),
            Retriever("static_value_46", DataType("s32")),  # always 0x17, now 0x2e (46)?
            Retriever("ai_script_goal", DataType("s32")),
            Retriever("aa_quantity", DataType("u8"),
                      set_repeat="1 if {effect_type} == 31 or {effect_type} == 28 else 0"),
            Retriever("aa_armor_or_attack_type", DataType("s24"),
                      set_repeat="1 if {effect_type} == 31 or {effect_type} == 28 else 0"),
            Retriever("quantity", DataType("s32"),
                      set_repeat="1 if {effect_type} != 31 and {effect_type} != 28 else 0"),
            Retriever("tribute_list", DataType("s32")),
            Retriever("diplomacy", DataType("s32")),
            Retriever("number_of_units_selected", DataType("s32"), save_as="number_of_units_selected"),
            Retriever("unknown", DataType("s32")),
            Retriever("object_list_unit_id", DataType("s32")),
            Retriever("player_source", DataType("s32")),
            Retriever("player_target", DataType("s32")),
            Retriever("technology", DataType("s32")),
            Retriever("string_id", DataType("s32")),
            Retriever("unknown_2", DataType("s32")),
            Retriever("display_time", DataType("s32")),
            Retriever("trigger_id", DataType("s32")),
            Retriever("location_x", DataType("s32")),
            Retriever("location_y", DataType("s32")),
            Retriever("area_1_x", DataType("s32")),
            Retriever("area_1_y", DataType("s32")),
            Retriever("area_2_x", DataType("s32")),
            Retriever("area_2_y", DataType("s32")),
            Retriever("object_group", DataType("s32")),
            Retriever("object_type", DataType("s32")),
            Retriever("instruction_panel_position", DataType("s32")),
            Retriever("attack_stance", DataType("s32")),
            Retriever("time_unit", DataType("s32")),
            Retriever("enabled_or_victory", DataType("s32")),
            Retriever("food", DataType("s32")),
            Retriever("wood", DataType("s32")),
            Retriever("stone", DataType("s32")),
            Retriever("gold", DataType("s32")),
            Retriever("item_id", DataType("s32")),
            Retriever("flash_object", DataType("s32")),
            Retriever("force_research_technology", DataType("s32")),
            Retriever("visibility_state", DataType("s32")),
            Retriever("scroll", DataType("s32")),
            Retriever("operation", DataType("s32")),
            Retriever("object_list_unit_id_2", DataType("s32")),
            Retriever("button_location", DataType("s32")),
            Retriever("ai_signal_value", DataType("s32")),
            Retriever("unknown_3", DataType("s32")),
            Retriever("object_attributes", DataType("s32")),
            Retriever("from_variable", DataType("s32")),
            Retriever("variable_or_timer", DataType("s32")),
            Retriever("facet", DataType("s32")),
            Retriever("unknown_4", DataType("s32")),
            Retriever("play_sound", DataType("s32")),
            Retriever("message", DataType("str32")),
            Retriever("sound_name", DataType("str32")),
            Retriever("selected_object_id", DataType("s32"), set_repeat="{number_of_units_selected}"),
        ]

        super().__init__("Effect", retrievers, parser_obj, data)



import pyautogui
from time import sleep

# [Condition] Accumulate Attribute
# distance_to_list = 9
# distance_to_copy = 4

for x in range(5, 0, -1):
    print(x, end=" ")
    sleep(1)
print("\n")

# Change to 'tabbing mode'
pyautogui.press('tab')
sleep(0.2)

for _ in range(0, 12):
    pyautogui.press('tab')

while True:
    pyautogui.press('tab')
    pyautogui.hotkey('shiftleft', 'tab')
    pyautogui.hotkey('shiftleft', 'tab')

    pyautogui.press(['1', '0'])

    for _ in range(0, 25):
        pyautogui.hotkey('tab')
    sleep(0.05)
    pyautogui.press('enter')
    sleep(0.05)
    for _ in range(0, 7):
        pyautogui.press('tab')

    pyautogui.press('enter')
    pyautogui.press('down')
    pyautogui.press('enter')



import abc


class AoE2Object:
    def __init__(self):
        pass

    @staticmethod
    @abc.abstractmethod
    def _parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    @abc.abstractmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):
        pass

    def __repr__(self):
        return str(self.__class__.__name__) + ": " + str(self.__dict__)



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.pieces.structs.terrain import TerrainStruct


class MapPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Separator??", DataType("2")),
            Retriever("Unknown String", DataType("str16")),
            Retriever("Separator?? (2)", DataType("2")),
            Retriever("Map color mood", DataType("str16")),
            Retriever("Collide and Correcting", DataType("u8")),
            # [VERSION CHANGE] ADDED in 1.36 > 1.37
            Retriever("Villager Force Drop", DataType("u8"), set_repeat="1 if '{scenario_version}' == '1.37' else 0"),
            Retriever("Player 1 camera, Y", DataType("s32")),
            Retriever("Player 1 camera, X", DataType("s32")),
            Retriever("AI Type", DataType("s8")),
            Retriever("Map Width", DataType("s32"), save_as="map_width"),
            Retriever("Map Height", DataType("s32"), save_as="map_height"),
            Retriever("Terrain data", DataType(TerrainStruct), set_repeat="{map_width}*{map_height}"),
        ]

        super().__init__("Map", retrievers, parser_obj, data=data)



from __future__ import annotations

from AoE2ScenarioParser.datasets import conditions
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.pieces.structs.condition import ConditionStruct


class ConditionObject(AoE2Object):
    def __init__(self,
                 condition_type,
                 amount_or_quantity,
                 resource_type_or_tribute_list,
                 unit_object,
                 next_object,
                 object_list,
                 player,
                 technology,
                 timer,
                 area_1_x,
                 area_1_y,
                 area_2_x,
                 area_2_y,
                 object_group,
                 object_type,
                 ai_signal,
                 inverted,
                 variable,
                 comparison,
                 target_player
                 ):

        self.condition_type = condition_type
        self.amount_or_quantity = amount_or_quantity
        self.resource_type_or_tribute_list = resource_type_or_tribute_list
        self.unit_object = unit_object
        self.next_object = next_object
        self.object_list = object_list
        self.player = player
        self.technology = technology
        self.timer = timer
        self.area_1_x = area_1_x
        self.area_1_y = area_1_y
        self.area_2_x = area_2_x
        self.area_2_y = area_2_y
        self.object_group = object_group
        self.object_type = object_type
        self.ai_signal = ai_signal
        self.inverted = inverted
        self.variable = variable
        self.comparison = comparison
        self.target_player = target_player

        super().__init__()

    def get_content_as_string(self):
        attributes_list = conditions.attributes[self.condition_type]

        return_string = ""
        for attribute in attributes_list:
            attr = getattr(self, attribute)
            if attribute == "condition_type" or attr == [] or attr == "" or attr == " " or attr == -1:
                continue
            return_string += "\t\t\t\t" + attribute + ": " + str(attr) + "\n"

        return return_string

    @staticmethod
    def _parse_object(parsed_data, **kwargs) -> ConditionObject:  # Expected {condition=conditionStruct}
        condition_struct = kwargs['condition']

        condition_type = find_retriever(condition_struct.retrievers, "condition_type").data
        parameters = conditions.attributes.get(condition_type)

        parameter_dict = conditions.empty_attributes.copy()
        for param in parameters:
            parameter_dict[param] = find_retriever(condition_struct.retrievers, param).data

        return ConditionObject(
            **parameter_dict
        )

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs) -> None:
        # Expected {condition=condition_obj, conditions=conditionsList}
        condition_obj = kwargs['condition']
        conditions_list = kwargs['conditions']

        data_list = [value for key, value in vars(condition_obj).items()]
        data_list.insert(1, 21)  # static_value_21
        data_list.insert(10, -1)  # unknown
        data_list.insert(19, -1)  # unknown_2
        conditions_list.append(ConditionStruct(data=data_list))



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class TerrainStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Terrain ID", DataType("u8")),
            Retriever("Elevation", DataType("u8")),
            Retriever("Unused", DataType("u8")),
            Retriever("Separator?", DataType("4"))
        ]

        super().__init__("Terrain", retrievers, parser_obj, data)



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType


class FileHeaderPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Version", DataType("c4"), save_as="scenario_version"),
            Retriever("Header length", DataType("u32")),
            Retriever("Savable", DataType("s32")),
            Retriever("Timestamp of last save", DataType("u32")),
            Retriever("Scenario instructions", DataType("str32")),
            Retriever("Individual victories used", DataType("u32")),
            Retriever("Player count", DataType("u32")),
            Retriever("Data", DataType("36")),
            Retriever("Creator name", DataType("str32")),
            Retriever("Trigger count", DataType("u32")),
        ]

        super().__init__("File Header", retrievers, parser_obj, data=data)



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class PlayerDataThreeStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Constant name", DataType("str16")),
            Retriever("Initial Camera, X", DataType("f32")),
            Retriever("Initial Camera, Y", DataType("f32")),
            Retriever("Unknown, similar to camera X", DataType("s16")),
            Retriever("Unknown, similar to camera Y", DataType("s16")),
            Retriever("[AOK] Allied Victory", DataType("u8")),
            Retriever("Player count for diplomacy", DataType("u16"), save_as="diplo_player_count"),
            Retriever("Diplomacy for interaction", DataType("u8"), set_repeat="{diplo_player_count}"),
            Retriever("Diplomacy for AI system", DataType("u32", repeat=9)),
            Retriever("Color", DataType("u32")),
            Retriever("Victory Version", DataType("f32"), save_as="vic_version"),
            Retriever("Unknown", DataType("u16"), save_as="unknown_DAT"),
            Retriever("Unknown (2)", DataType("u8"), set_repeat="8 if {vic_version} == 2 else 0"),
            Retriever("Unknown structure, (Grand Theft Empires)", DataType("44"), set_repeat="{unknown_DAT}"),
            Retriever("Unknown (3)", DataType("u8", repeat=7)),
            Retriever("Unknown (4)", DataType("s32")),
        ]

        super().__init__("Player Data #3", retrievers, parser_obj, data)



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class AIStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Unknown, always 0", DataType("u32")),
            Retriever("Unknown, always 0 (2)", DataType("u32")),
            Retriever("AI .per file text", DataType("str32")),
        ]

        super().__init__("AI", retrievers, parser_obj, data)



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class MapObject(AoE2Object):
    def __init__(self,
                 map_color_mood,
                 collide_and_correct,
                 map_width,
                 map_height,
                 terrain
                 ):

        self.map_color_mood = map_color_mood
        self.collide_and_correct = collide_and_correct
        self.map_width = map_width
        self.map_height = map_height
        self.terrain = terrain

        super().__init__()

    @staticmethod
    def _parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):
        pass




def create_generator(uncompressed_file, chunk_size=4):
    for x in range(0, len(uncompressed_file), chunk_size):
        yield uncompressed_file[x:x + chunk_size]


def repeat_generator(generator, run_times, return_bytes=True):
    elements = b'' if return_bytes else []

    for i in range(0, run_times):
        if return_bytes:
            elements += next(generator)
        else:
            # Generator returns list. So lists are merged
            elements += next(generator)

    return elements



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="AoE2ScenarioParser",
    version="0.0.7",
    author="Kerwin Sneijders",
    author_email="ksneijders@hotmail.com",
    description="This is a project for editing parts of an 'aoe2scenario' file from Age of Empires 2 Definitive Edition",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KSneijders/AoE2ScenarioParser",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'bdict'
    ]
)



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class MessagesObject(AoE2Object):
    def __init__(self,
                 instructions,
                 hints,
                 victory,
                 loss,
                 history,
                 scouts,
                 ascii_instructions,
                 ascii_hints,
                 ascii_victory,
                 ascii_loss,
                 ascii_history,
                 ascii_scouts,
                 ):

        self.instructions = instructions
        self.hints = hints
        self.victory = victory
        self.loss = loss
        self.history = history
        self.scouts = scouts
        self.ascii_instructions = ascii_instructions
        self.ascii_hints = ascii_hints
        self.ascii_victory = ascii_victory
        self.ascii_loss = ascii_loss
        self.ascii_history = ascii_history
        self.ascii_scouts = ascii_scouts

        super().__init__()

    @staticmethod
    def _parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):
        pass



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.pieces.structs.player_data_four import PlayerDataFourStruct
from AoE2ScenarioParser.pieces.structs.player_data_three_struct import PlayerDataThreeStruct
from AoE2ScenarioParser.pieces.structs.player_units import PlayerUnitsStruct


class UnitsPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Number of unit sections", DataType("u32"), save_as="unit_sections"),  # Always 9 (Gaia + 8Plyrs)
            Retriever("Player data #4", DataType(PlayerDataFourStruct, repeat=8)),
            Retriever("Number of players?", DataType("u32")),  # Also always 9 (Gaia + 8Plyrs)
            Retriever("Player data #3", DataType(PlayerDataThreeStruct, repeat=8)),
            Retriever("Player Units", DataType(PlayerUnitsStruct), set_repeat="{unit_sections}"),
        ]

        super().__init__("Units", retrievers, parser_obj, data=data)



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class PlayerObject(AoE2Object):
    def __init__(self,
                 player_number,
                 active,
                 human,
                 civilization,
                 gold,
                 wood,
                 food,
                 stone,
                 color,
                 starting_age,
                 pop_limit
                 ):

        self.player_number = player_number
        self.active = active
        self.human = human
        self.civilization = civilization
        self.gold = gold
        self.wood = wood
        self.food = food
        self.stone = stone
        self.color = color
        self.starting_age = starting_age
        self.pop_limit = pop_limit

        super().__init__()

    @staticmethod
    def _parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):
        pass



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class TerrainObject(AoE2Object):
    def __init__(self,
                 terrain_id,
                 elevation
                 ):
        self.terrain_id = terrain_id
        self.elevation = elevation

        super().__init__()

    @staticmethod
    def _parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):
        pass



from AoE2ScenarioParser.datasets import effects, conditions
from AoE2ScenarioParser.datasets.conditions import Condition
from AoE2ScenarioParser.datasets.effects import Effect
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.condition_obj import ConditionObject
from AoE2ScenarioParser.objects.effect_obj import EffectObject
from AoE2ScenarioParser.pieces.structs.trigger import TriggerStruct


class TriggerObject(AoE2Object):
    def __init__(self,
                 name,
                 description="",
                 description_stid=-1,
                 display_as_objective=0,
                 short_description="",
                 short_description_stid=-1,
                 display_on_screen=0,
                 description_order=0,
                 enabled=1,
                 looping=0,
                 header=0,
                 mute_objectives=0,
                 conditions_list=None,
                 condition_order=None,
                 effects_list=None,
                 effect_order=None,
                 trigger_id=-1,
                 ):

        if conditions_list is None:
            conditions_list = []
        if condition_order is None:
            condition_order = []
        if effects_list is None:
            effects_list = []
        if effect_order is None:
            effect_order = []

        self.name = name
        self.description = description
        self.description_stid = description_stid
        self.display_as_objective = display_as_objective
        self.short_description = short_description
        self.short_description_stid = short_description_stid
        self.display_on_screen = display_on_screen
        self.description_order = description_order
        self.enabled = enabled
        self.looping = looping
        self.header = header
        self.mute_objectives = mute_objectives
        self.conditions = conditions_list
        self.condition_order = condition_order
        self.effects = effects_list
        self.effect_order = effect_order
        self.trigger_id = trigger_id

        self.trigger_id = trigger_id

        super().__init__()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = helper.add_str_trail(val)

    @property
    def description(self):
        return self.__description

    @description.setter
    def description(self, val):
        self.__description = helper.add_str_trail(val)

    @property
    def short_description(self):
        return self.__short_description

    @short_description.setter
    def short_description(self, val):
        self.__short_description = helper.add_str_trail(val)

    def add_effect(self, effect_type: Effect):
        new_effect = EffectObject(**effects.default_attributes[effect_type.value])
        self.effects.append(new_effect)
        helper.update_order_array(self.effect_order, len(self.effects))
        return new_effect

    def add_condition(self, condition_type: Condition):
        new_cond = ConditionObject(**conditions.default_attributes[condition_type.value])
        self.conditions.append(new_cond)
        helper.update_order_array(self.condition_order, len(self.conditions))
        return new_cond

    def get_content_as_string(self):
        return_string = ""
        data_tba = [
            ('enabled', self.enabled != 0),
            ('looping', self.looping != 0)
        ]

        if helper.del_str_trail(self.description) != "":
            data_tba.append(('description', "'" + helper.del_str_trail(self.description) + "'"))
        if self.description_stid != -1:
            data_tba.append(('description_stid', self.description_stid))
        if helper.del_str_trail(self.short_description) != "":
            data_tba.append(('short_description', "'" + helper.del_str_trail(self.short_description) + "'"))
        if self.short_description_stid != -1:
            data_tba.append(('short_description_stid', self.short_description_stid))
        if self.display_as_objective != 0:
            data_tba.append(('display_as_objective', self.display_as_objective != 0))
        if self.display_on_screen != 0:
            data_tba.append(('display_on_screen', self.display_on_screen != 0))
        if self.description_order != 0:
            data_tba.append(('description_order', self.description_order))
        if self.header != 0:
            data_tba.append(('header', self.header != 0))
        if self.mute_objectives != 0:
            data_tba.append(('mute_objectives', self.mute_objectives != 0))

        for data in data_tba:
            return_string += "\t\t" + data[0] + ": " + str(data[1]) + "\n"

        if len(self.condition_order) > 0:
            return_string += "\t\tconditions:\n"
            for c_display_order, condition_id in enumerate(self.condition_order):
                condition = self.conditions[condition_id]

                return_string += "\t\t\t" + conditions.condition_names[condition.condition_type] + ":\n"
                return_string += condition.get_content_as_string()

        if len(self.effect_order) > 0:
            return_string += "\t\teffects:\n"
            for e_display_order, effect_id in enumerate(self.effect_order):
                effect = self.effects[effect_id]

                return_string += "\t\t\t" + effects.effect_names[effect.effect_type] + ":\n"
                return_string += effect.get_content_as_string()

        return return_string

    @staticmethod
    def _parse_object(parsed_data, **kwargs):  # Expected {trigger=triggerStruct, trigger_id=id}
        trigger = kwargs['trigger']

        effects_list = []
        effect_structs = parser.listify(find_retriever(trigger.retrievers, "Effect data").data)
        for effect_struct in effect_structs:
            effects_list.append(EffectObject._parse_object(parsed_data, effect=effect_struct))

        conditions_list = []
        condition_structs = parser.listify(find_retriever(trigger.retrievers, "Condition data").data)
        for condition_struct in condition_structs:
            conditions_list.append(ConditionObject._parse_object(parsed_data, condition=condition_struct))

        return TriggerObject(
            name=find_retriever(trigger.retrievers, "Trigger name").data,
            description=find_retriever(trigger.retrievers, "Trigger description").data,
            description_stid=find_retriever(trigger.retrievers, "Description string Table ID").data,
            display_as_objective=find_retriever(trigger.retrievers, "Act as objective").data,
            short_description=find_retriever(trigger.retrievers, "Short description").data,
            short_description_stid=find_retriever(trigger.retrievers, "Short description string Table ID").data,
            display_on_screen=find_retriever(trigger.retrievers, "Display on screen").data,
            description_order=find_retriever(trigger.retrievers, "Description order (in objectives)").data,
            enabled=find_retriever(trigger.retrievers, "Enabled").data,
            looping=find_retriever(trigger.retrievers, "Looping").data,
            header=find_retriever(trigger.retrievers, "Make header").data,
            mute_objectives=find_retriever(trigger.retrievers, "Mute objectives").data,
            conditions_list=conditions_list,
            condition_order=parser.listify(find_retriever(trigger.retrievers, "Condition display order array").data),
            effects_list=effects_list,
            effect_order=parser.listify(find_retriever(trigger.retrievers, "Effect display order array").data),
            trigger_id=kwargs['trigger_id'],
        )

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):  # Expected {trigger=triggerStruct}
        trigger_data_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger data")
        trigger = kwargs['trigger']
        trigger.effect_order = parser.listify(trigger.effect_order)
        trigger.condition_order = parser.listify(trigger.condition_order)

        effects_list = []
        for effect_obj in trigger.effects:
            EffectObject._reconstruct_object(parsed_header, parsed_data, objects, effect=effect_obj,
                                             effects=effects_list)

        helper.update_order_array(trigger.effect_order, len(trigger.effects))

        conditions_list = []
        for condition_obj in trigger.conditions:
            ConditionObject._reconstruct_object(parsed_header, parsed_data, objects, condition=condition_obj,
                                                conditions=conditions_list)

        helper.update_order_array(trigger.condition_order, len(trigger.conditions))

        trigger_data_retriever.data.append(
            TriggerStruct(data=[
                trigger.enabled,
                trigger.looping,
                trigger.description_stid,
                trigger.display_as_objective,
                trigger.description_order,
                trigger.header,
                trigger.short_description_stid,
                trigger.display_on_screen,
                b'\x00\x00\x00\x00\x00',  # Unknown
                trigger.mute_objectives,
                trigger.description,
                trigger.name,
                trigger.short_description,
                len(trigger.effects),
                effects_list,
                trigger.effect_order,
                len(trigger.conditions),
                conditions_list,
                trigger.condition_order,
            ])
        )



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class UnitStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("X position", DataType("f32")),
            Retriever("Y position", DataType("f32")),
            Retriever("Z position", DataType("f32")),
            Retriever("ID", DataType("u32")),
            Retriever("Unit 'constant'", DataType("u16")),
            Retriever("Status", DataType("u8")),
            # Status, Always 2. 0-6 no difference (?) | 7-255 makes it disappear. (Except from the mini-map)
            Retriever("Rotation, in radians", DataType("f32")),
            Retriever("Initial animation frame", DataType("u16")),
            Retriever("Garrisoned in: ID", DataType("s32")),
        ]

        super().__init__("Unit", retrievers, parser_obj, data)



from __future__ import annotations

from enum import Enum
from typing import List

from AoE2ScenarioParser.datasets import effects
from AoE2ScenarioParser.datasets.techs import Tech
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.pieces.structs.effect import EffectStruct


class EffectObject(AoE2Object):
    def __init__(self,
                 effect_type,
                 ai_script_goal,
                 aa_quantity,
                 aa_armor_or_attack_type,
                 quantity,
                 tribute_list,
                 diplomacy,
                 number_of_units_selected,
                 object_list_unit_id,
                 player_source,
                 player_target,
                 technology,
                 string_id,
                 display_time,
                 trigger_id,
                 location_x,
                 location_y,
                 area_1_x,
                 area_1_y,
                 area_2_x,
                 area_2_y,
                 object_group,
                 object_type,
                 instruction_panel_position,
                 attack_stance,
                 time_unit,
                 enabled_or_victory,
                 food,
                 wood,
                 stone,
                 gold,
                 item_id,
                 flash_object,
                 force_research_technology,
                 visibility_state,
                 scroll,
                 operation,
                 object_list_unit_id_2,
                 button_location,
                 ai_signal_value,
                 object_attributes,
                 from_variable,
                 variable_or_timer,
                 facet,
                 play_sound,
                 message="",
                 sound_name="",
                 selected_object_id: List = None,
                 ):

        if selected_object_id is None:
            selected_object_id = []

        self.effect_type = effect_type
        self.ai_script_goal = ai_script_goal
        self.aa_quantity = aa_quantity
        self.aa_armor_or_attack_type = aa_armor_or_attack_type
        self.quantity = quantity
        self.tribute_list = tribute_list
        self.diplomacy = diplomacy
        self.number_of_units_selected = number_of_units_selected
        self.object_list_unit_id = object_list_unit_id
        self.player_source = player_source
        self.player_target = player_target
        self.technology = technology
        self.string_id = string_id
        self.display_time = display_time
        self.trigger_id = trigger_id
        self.location_x = location_x
        self.location_y = location_y
        self.area_1_x = area_1_x
        self.area_1_y = area_1_y
        self.area_2_x = area_2_x
        self.area_2_y = area_2_y
        self.object_group = object_group
        self.object_type = object_type
        self.instruction_panel_position = instruction_panel_position
        self.attack_stance = attack_stance
        self.time_unit = time_unit
        self.enabled_or_victory = enabled_or_victory
        self.food = food
        self.wood = wood
        self.stone = stone
        self.gold = gold
        self.item_id = item_id
        self.flash_object = flash_object
        self.force_research_technology = force_research_technology
        self.visibility_state = visibility_state
        self.scroll = scroll
        self.operation = operation
        self.object_list_unit_id_2 = object_list_unit_id_2
        self.button_location = button_location
        self.ai_signal_value = ai_signal_value
        self.object_attributes = object_attributes
        self.from_variable = from_variable
        self.variable_or_timer = variable_or_timer
        self.facet = facet
        self.play_sound = play_sound
        self.message = message
        self.sound_name = sound_name
        self.selected_object_id = selected_object_id

        super().__init__()

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, val):
        self._message = helper.add_str_trail(val)

    @property
    def sound_name(self):
        return self._sound_name

    @sound_name.setter
    def sound_name(self, val):
        self._sound_name = helper.add_str_trail(val)

    def get_content_as_string(self):
        attributes_list = effects.attributes[self.effect_type]

        return_string = ""
        for attribute in attributes_list:
            attr = getattr(self, attribute)
            if attribute == "effect_type" or attr == [] or attr == "" or attr == " " or attr == -1:
                continue
            return_string += "\t\t\t\t" + attribute + ": " + str(attr) + "\n"

        if return_string == "":
            return "\t\t\t\t< no attributes >\n"

        return return_string

    @staticmethod
    def _parse_object(parsed_data, **kwargs):  # Expected {effect=effectStruct}
        effect_struct = kwargs['effect']

        effect_type = find_retriever(effect_struct.retrievers, "effect_type").data
        parameters = effects.attributes.get(effect_type)

        parameter_dict = effects.empty_attributes.copy()
        for param in parameters:
            parameter_dict[param] = find_retriever(effect_struct.retrievers, param).data

        return EffectObject(
            **parameter_dict
        )

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):  # Expected {effect=effect_obj, effects=effectsList}
        effect_obj = kwargs['effect']
        effects_list = kwargs['effects']

        data_list = [value for key, value in vars(effect_obj).items()]
        data_list.insert(1, 46)  # static_value_46
        data_list.insert(9, -1)  # unknown
        data_list.insert(15, -1)  # unknown_2
        data_list.insert(43, -1)  # unknown_3
        data_list.insert(48, -1)  # unknown_4

        effects_list.append(EffectStruct(data=data_list))



from enum import IntEnum

from bidict import bidict


class Building(IntEnum):
    AACHEN_CATHEDRAL = 1622
    AMPHITHEATRE = 251
    AQUEDUCT = 231
    ARCH_OF_CONSTANTINE = 899
    ARCHERY_RANGE = 87
    ARMY_TENT_A = 1196
    ARMY_TENT_B = 1197
    ARMY_TENT_C = 1198
    ARMY_TENT_D = 1199
    ARMY_TENT_E = 1200
    BARRACKS = 12
    BLACKSMITH = 103
    BOMBARD_TOWER = 236
    BRIDGE_A_BOTTOM = 607
    BRIDGE_A_BROKEN_BOTTOM = 740
    BRIDGE_A_BROKEN_TOP = 739
    BRIDGE_A_CRACKED = 738
    BRIDGE_A_MIDDLE = 606
    BRIDGE_A_TOP = 605
    BRIDGE_B_BOTTOM = 610
    BRIDGE_B_BROKEN_BOTTOM = 743
    BRIDGE_B_BROKEN_TOP = 742
    BRIDGE_B_CRACKED = 741
    BRIDGE_B_MIDDLE = 609
    BRIDGE_B_TOP = 608
    BRIDGE_C_BOTTOM = 1206
    BRIDGE_C_BROKEN_BOTTOM = 1212
    BRIDGE_C_BROKEN_TOP = 1211
    BRIDGE_C_CRACKED = 1210
    BRIDGE_C_MIDDLE = 1205
    BRIDGE_C_TOP = 1204
    BRIDGE_D_BOTTOM = 1209
    BRIDGE_D_BROKEN_BOTTOM = 1215
    BRIDGE_D_BROKEN_TOP = 1214
    BRIDGE_D_CRACKED = 1213
    BRIDGE_D_MIDDLE = 1208
    BRIDGE_D_TOP = 1207
    BRIDGE_E_BOTTOM = 1552
    BRIDGE_E_BROKEN_BOTTOM = 1558
    BRIDGE_E_BROKEN_TOP = 1557
    BRIDGE_E_CRACKED = 1556
    BRIDGE_E_MIDDLE = 1551
    BRIDGE_E_TOP = 1550
    BRIDGE_F_BOTTOM = 1555
    BRIDGE_F_BROKEN_BOTTOM = 1561
    BRIDGE_F_BROKEN_TOP = 1560
    BRIDGE_F_CRACKED = 1559
    BRIDGE_F_MIDDLE = 1554
    BRIDGE_F_TOP = 1553
    CASTLE = 82
    CATHEDRAL = 599
    CHAIN_WEST_TO_EAST = 1398
    CHAIN_SOUTHWEST_TO_NORTHEAST = 1396
    CHAIN_NORTH_TO_SOUTH = 1399
    CHAIN_NORTHWEST_TO_SOUTHEAST = 1397
    CITY_GATE_WEST_TO_EAST = 1587
    CITY_GATE_SOUTHWEST_TO_NORTHEAST = 1579
    CITY_GATE_NORTH_TO_SOUTH = 1591
    CITY_GATE_NORTHWEST_TO_SOUTHEAST = 1583
    CITY_WALL = 370
    COLOSSEUM = 263
    DOCK = 45
    DORMITION_CATHEDRAL = 1369
    FARM = 50
    FEITORIA = 1021
    FENCE = 1062
    FIRE_TOWER = 190
    FISH_TRAP = 199
    FORTIFIED_PALISADE_WALL = 119
    FORTIFIED_TOWER = 1102
    FORTIFIED_WALL = 155
    FORTRESS = 33
    GATE_NORTHWEST_TO_SOUTHEAST = 88
    GATE_WEST_TO_EAST = 659
    GATE_SOUTHWEST_TO_NORTHEAST = 64
    GATE_NORTH_TO_SOUTH = 667
    GOL_GUMBAZ = 1217
    GUARD_TOWER = 234
    HARBOR = 1189
    HOUSE = 70
    HUT_A = 1082
    HUT_B = 1083
    HUT_C = 1084
    HUT_D = 1085
    HUT_E = 1086
    HUT_F = 1087
    HUT_G = 1088
    KEEP = 235
    KREPOST = 1251
    LUMBER_CAMP = 562
    MARKET = 84
    MILL = 68
    MINING_CAMP = 584
    MONASTERY = 104
    MONUMENT = 826
    OUTPOST = 598
    PALISADE_GATE_SOUTHWEST_TO_NORTHEAST = 793
    PALISADE_GATE_WEST_TO_EAST = 797
    PALISADE_GATE_NORTHWEST_TO_SOUTHEAST = 789
    PALISADE_GATE_NORTH_TO_SOUTH = 801
    PALISADE_WALL = 72
    PYRAMID = 689
    QUIMPER_CATHEDRAL = 872
    RICE_FARM = 1187
    ROCK_CHURCH = 1378
    SANCHI_STUPA = 1216
    SANKORE_MADRASAH = 1367
    SEA_GATE_SOUTHWEST_TO_NORTHEAST = 1379
    SEA_GATE_NORTH_TO_SOUTH = 1391
    SEA_GATE_WEST_TO_EAST = 1387
    SEA_GATE_NORTHWEST_TO_SOUTHEAST = 1383
    SEA_TOWER = 785
    SEA_WALL = 788
    SHRINE = 1264
    SIEGE_WORKSHOP = 49
    STABLE = 101
    STONE_WALL = 117
    STORAGE = 1081
    TEMPLE_OF_HEAVEN = 637
    TENT_A = 1097
    TENT_B = 1098
    TENT_C = 1099
    TENT_D = 1100
    TENT_E = 1101
    TOWER_OF_LONDON = 1368
    TOWN_CENTER = 109
    KHOSRAU = 444
    TRADE_WORKSHOP = 110
    UNIVERSITY = 209
    WATCH_TOWER = 79
    WONDER = 276
    WOODEN_BRIDGE_A_BOTTOM = 1311
    WOODEN_BRIDGE_A_MIDDLE = 1310
    WOODEN_BRIDGE_A_TOP = 1309
    WOODEN_BRIDGE_B_BOTTOM = 1314
    WOODEN_BRIDGE_B_MIDDLE = 1313
    WOODEN_BRIDGE_B_TOP = 1312
    YURT_A = 712
    YURT_B = 713
    YURT_C = 714
    YURT_D = 715
    YURT_E = 716
    YURT_F = 717
    YURT_G = 718
    YURT_H = 719


class GaiaBuilding(IntEnum):
    # Gaia Buildings
    RUINS = 345

    # Normal Buildings
    AACHEN_CATHEDRAL = 1622
    AMPHITHEATRE = 251
    AQUEDUCT = 231
    ARCH_OF_CONSTANTINE = 899
    ARCHERY_RANGE = 87
    ARMY_TENT_A = 1196
    ARMY_TENT_B = 1197
    ARMY_TENT_C = 1198
    ARMY_TENT_D = 1199
    ARMY_TENT_E = 1200
    BARRACKS = 12
    BLACKSMITH = 103
    BOMBARD_TOWER = 236
    BRIDGE_A_BOTTOM = 607
    BRIDGE_A_BROKEN_BOTTOM = 740
    BRIDGE_A_BROKEN_TOP = 739
    BRIDGE_A_CRACKED = 738
    BRIDGE_A_MIDDLE = 606
    BRIDGE_A_TOP = 605
    BRIDGE_B_BOTTOM = 610
    BRIDGE_B_BROKEN_BOTTOM = 743
    BRIDGE_B_BROKEN_TOP = 742
    BRIDGE_B_CRACKED = 741
    BRIDGE_B_MIDDLE = 609
    BRIDGE_B_TOP = 608
    BRIDGE_C_BOTTOM = 1206
    BRIDGE_C_BROKEN_BOTTOM = 1212
    BRIDGE_C_BROKEN_TOP = 1211
    BRIDGE_C_CRACKED = 1210
    BRIDGE_C_MIDDLE = 1205
    BRIDGE_C_TOP = 1204
    BRIDGE_D_BOTTOM = 1209
    BRIDGE_D_BROKEN_BOTTOM = 1215
    BRIDGE_D_BROKEN_TOP = 1214
    BRIDGE_D_CRACKED = 1213
    BRIDGE_D_MIDDLE = 1208
    BRIDGE_D_TOP = 1207
    BRIDGE_E_BOTTOM = 1552
    BRIDGE_E_BROKEN_BOTTOM = 1558
    BRIDGE_E_BROKEN_TOP = 1557
    BRIDGE_E_CRACKED = 1556
    BRIDGE_E_MIDDLE = 1551
    BRIDGE_E_TOP = 1550
    BRIDGE_F_BOTTOM = 1555
    BRIDGE_F_BROKEN_BOTTOM = 1561
    BRIDGE_F_BROKEN_TOP = 1560
    BRIDGE_F_CRACKED = 1559
    BRIDGE_F_MIDDLE = 1554
    BRIDGE_F_TOP = 1553
    CASTLE = 82
    CATHEDRAL = 599
    CHAIN_WEST_TO_EAST = 1398
    CHAIN_SOUTHWEST_TO_NORTHEAST = 1396
    CHAIN_NORTH_TO_SOUTH = 1399
    CHAIN_NORTHWEST_TO_SOUTHEAST = 1397
    CITY_GATE_WEST_TO_EAST = 1587
    CITY_GATE_SOUTHWEST_TO_NORTHEAST = 1579
    CITY_GATE_NORTH_TO_SOUTH = 1591
    CITY_GATE_NORTHWEST_TO_SOUTHEAST = 1583
    CITY_WALL = 370
    COLOSSEUM = 263
    DOCK = 45
    DORMITION_CATHEDRAL = 1369
    FARM = 50
    FEITORIA = 1021
    FENCE = 1062
    FIRE_TOWER = 190
    FISH_TRAP = 199
    FORTIFIED_PALISADE_WALL = 119
    FORTIFIED_TOWER = 1102
    FORTIFIED_WALL = 155
    FORTRESS = 33
    GATE_NORTHWEST_TO_SOUTHEAST = 88
    GATE_WEST_TO_EAST = 659
    GATE_SOUTHWEST_TO_NORTHEAST = 64
    GATE_NORTH_TO_SOUTH = 667
    GOL_GUMBAZ = 1217
    GUARD_TOWER = 234
    HARBOR = 1189
    HOUSE = 70
    HUT_A = 1082
    HUT_B = 1083
    HUT_C = 1084
    HUT_D = 1085
    HUT_E = 1086
    HUT_F = 1087
    HUT_G = 1088
    KEEP = 235
    KREPOST = 1251
    LUMBER_CAMP = 562
    MARKET = 84
    MILL = 68
    MINING_CAMP = 584
    MONASTERY = 104
    MONUMENT = 826
    OUTPOST = 598
    PALISADE_GATE_SOUTHWEST_TO_NORTHEAST = 793
    PALISADE_GATE_WEST_TO_EAST = 797
    PALISADE_GATE_NORTHWEST_TO_SOUTHEAST = 789
    PALISADE_GATE_NORTH_TO_SOUTH = 801
    PALISADE_WALL = 72
    PYRAMID = 689
    QUIMPER_CATHEDRAL = 872
    RICE_FARM = 1187
    ROCK_CHURCH = 1378
    SANCHI_STUPA = 1216
    SANKORE_MADRASAH = 1367
    SEA_GATE_SOUTHWEST_TO_NORTHEAST = 1379
    SEA_GATE_NORTH_TO_SOUTH = 1391
    SEA_GATE_WEST_TO_EAST = 1387
    SEA_GATE_NORTHWEST_TO_SOUTHEAST = 1383
    SEA_TOWER = 785
    SEA_WALL = 788
    SHRINE = 1264
    SIEGE_WORKSHOP = 49
    STABLE = 101
    STONE_WALL = 117
    STORAGE = 1081
    TEMPLE_OF_HEAVEN = 637
    TENT_A = 1097
    TENT_B = 1098
    TENT_C = 1099
    TENT_D = 1100
    TENT_E = 1101
    TOWER_OF_LONDON = 1368
    TOWN_CENTER = 109
    KHOSRAU = 444
    TRADE_WORKSHOP = 110
    UNIVERSITY = 209
    WATCH_TOWER = 79
    WONDER = 276
    WOODEN_BRIDGE_A_BOTTOM = 1311
    WOODEN_BRIDGE_A_MIDDLE = 1310
    WOODEN_BRIDGE_A_TOP = 1309
    WOODEN_BRIDGE_B_BOTTOM = 1314
    WOODEN_BRIDGE_B_MIDDLE = 1313
    WOODEN_BRIDGE_B_TOP = 1312
    YURT_A = 712
    YURT_B = 713
    YURT_C = 714
    YURT_D = 715
    YURT_E = 716
    YURT_F = 717
    YURT_G = 718
    YURT_H = 719


building_names = bidict({
    12: "barracks",
    33: "fortress",
    45: "dock",
    49: "siege_workshop",
    50: "farm",
    64: "gate_southwest_to_northeast",
    68: "mill",
    70: "house",
    72: "palisade_wall",
    79: "watch_tower",
    82: "castle",
    84: "market",
    87: "archery_range",
    88: "gate_northwest_to_southeast",
    101: "stable",
    103: "blacksmith",
    104: "monastery",
    109: "town_center",
    110: "trade_workshop",
    117: "stone_wall",
    119: "fortified_palisade_wall",
    155: "fortified_wall",
    190: "fire_tower",
    199: "fish_trap",
    209: "university",
    231: "aqueduct",
    234: "guard_tower",
    235: "keep",
    236: "bombard_tower",
    251: "amphitheatre",
    263: "colosseum",
    276: "wonder",
    370: "city_wall",
    444: "khosrau",
    562: "lumber_camp",
    584: "mining_camp",
    598: "outpost",
    599: "cathedral",
    605: "bridge_a_top",
    606: "bridge_a_middle",
    607: "bridge_a_bottom",
    608: "bridge_b_top",
    609: "bridge_b_middle",
    610: "bridge_b_bottom",
    637: "temple_of_heaven",
    659: "gate_west_to_east",
    667: "gate_north_to_south",
    689: "pyramid",
    712: "yurt_a",
    713: "yurt_b",
    714: "yurt_c",
    715: "yurt_d",
    716: "yurt_e",
    717: "yurt_f",
    718: "yurt_g",
    719: "yurt_h",
    738: "bridge_a_cracked",
    739: "bridge_a_broken_top",
    740: "bridge_a_broken_bottom",
    741: "bridge_b_cracked",
    742: "bridge_b_broken_top",
    743: "bridge_b_broken_bottom",
    785: "sea_tower",
    788: "sea_wall",
    793: "palisade_gate_southwest_to_northeast",
    797: "palisade_gate_west_to_east",
    789: "palisade_gate_northwest_to_southeast",
    801: "palisade_gate_north_to_south",
    826: "monument",
    872: "quimper_cathedral",
    899: "arch_of_constantine",
    1021: "feitoria",
    1062: "fence",
    1081: "storage",
    1082: "hut_a",
    1083: "hut_b",
    1084: "hut_c",
    1085: "hut_d",
    1086: "hut_e",
    1087: "hut_f",
    1088: "hut_g",
    1097: "tent_a",
    1098: "tent_b",
    1099: "tent_c",
    1100: "tent_d",
    1101: "tent_e",
    1102: "fortified_tower",
    1187: "rice_farm",
    1189: "harbor",
    1196: "army_tent_a",
    1197: "army_tent_b",
    1198: "army_tent_c",
    1199: "army_tent_d",
    1200: "army_tent_e",
    1204: "bridge_c_top",
    1205: "bridge_c_middle",
    1206: "bridge_c_bottom",
    1207: "bridge_d_top",
    1208: "bridge_d_middle",
    1209: "bridge_d_bottom",
    1210: "bridge_c_cracked",
    1211: "bridge_c_broken_top",
    1212: "bridge_c_broken_bottom",
    1213: "bridge_d_cracked",
    1214: "bridge_d_broken_top",
    1215: "bridge_d_broken_bottom",
    1216: "sanchi_stupa",
    1217: "gol_gumbaz",
    1251: "krepost",
    1264: "shrine",
    1309: "wooden_bridge_a_top",
    1310: "wooden_bridge_a_middle",
    1311: "wooden_bridge_a_bottom",
    1312: "wooden_bridge_b_top",
    1313: "wooden_bridge_b_middle",
    1314: "wooden_bridge_b_bottom",
    1367: "sankore_madrasah",
    1368: "tower_of_london",
    1369: "dormition_cathedral",
    1378: "rock_church",
    1379: "sea_gate_southwest_to_northeast",
    1391: "sea_gate_north_to_south",
    1387: "sea_gate_west_to_east",
    1383: "sea_gate_northwest_to_southeast",
    1396: "chain_southwest_to_northeast",
    1397: "chain_northwest_to_southeast",
    1398: "chain_west_to_east",
    1399: "chain_north_to_south",
    1550: "bridge_e_top",
    1551: "bridge_e_middle",
    1552: "bridge_e_bottom",
    1553: "bridge_f_top",
    1554: "bridge_f_middle",
    1555: "bridge_f_bottom",
    1556: "bridge_e_cracked",
    1557: "bridge_e_broken_top",
    1558: "bridge_e_broken_bottom",
    1559: "bridge_f_cracked",
    1560: "bridge_f_broken_top",
    1561: "bridge_f_broken_bottom",
    1579: "city_gate_southwest_to_northeast",
    1583: "city_gate_northwest_to_southeast",
    1587: "city_gate_west_to_east",
    1591: "city_gate_north_to_south",
    1622: "aachen_cathedral",

    # Gaia Buildings
    345: "ruins",
})



from AoE2ScenarioParser.helper.generator import create_generator
from AoE2ScenarioParser.pieces import aoe2_piece


class AoE2Struct(aoe2_piece.AoE2Piece):
    def __init__(self, piece_type, retrievers, parser_obj=None, data=None):
        super().__init__(piece_type, retrievers, parser_obj, data)

        if data and parser_obj:
            gen = create_generator(data, 1)
            super().set_data_from_generator(gen)

    def _entry_to_string(self, name, data, datatype):
        return "\t\t\t" + name + ": " + data + " (" + datatype + ")\n"

    def get_header_string(self):
        return "############ " + self.piece_type + " ############"

    def __str__(self):
        """ Remove Terrain and EyeCandy Units from the __str__ representation. As it'd mostly be considered spam. """
        if self.piece_type in ["Terrain"]:
            return "."  # Recommended to keep as '.' due to the amount of tiles in a map (Tiny map = 14400) or units.
        elif self.piece_type in ["Unit"]:
            if self.retrievers[4].data == 1358:  # Eye Candy (eg. Fake plants)
                return "."
        return super().__str__()



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.ai import AIStruct
from AoE2ScenarioParser.pieces.structs.resources import ResourcesStruct


class PlayerDataTwoPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Strings", DataType("str16", repeat=32)),
            Retriever("AI names", DataType("str16", repeat=16)),
            Retriever("AI files", DataType(AIStruct, repeat=16)),
            Retriever("AI type", DataType("u8", repeat=16)),
            Retriever("Separator", DataType("u32")),
            Retriever("Resources", DataType(ResourcesStruct, repeat=16))
        ]

        super().__init__("Player Data #2", retrievers, parser_obj, data=data)



from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.retriever import find_retriever


class AoE2Piece:
    def __init__(self, piece_type, retrievers, parser_obj=None, data=None):
        self.piece_type = piece_type
        self.retrievers = retrievers
        self.parser = parser_obj
        if data:
            self.set_data(data)

    def set_data(self, data):
        saves = {}
        if len(data) == len(self.retrievers):
            for i in range(0, len(data)):
                if self.retrievers[i].set_repeat:
                    self.retrievers[i].datatype.repeat = parser.parse_repeat_string(
                        saves,
                        self.retrievers[i].set_repeat
                    )

                if self.retrievers[i].log_value:
                    print(self.retrievers[i], "was set to:", parser.vorl(data[i]))
                self.retrievers[i].set_data(data[i])

                if self.retrievers[i].save_as is not None:
                    if type(data[i]) is not list:
                        data[i] = [data[i]]
                    saves[self.retrievers[i].save_as] = parser.vorl(data[i])
        else:
            ValueError("Data list isn't the same size as the DataType list")

    def get_value(self, retriever_key):
        return find_retriever(self.retrievers, retriever_key).data

    def get_length(self):
        total_length = 0
        try:
            for i in range(0, len(self.retrievers)):
                datatype, length = parser.datatype_to_type_length(self.retrievers[i].datatype.var)

                if datatype == "struct":
                    if type(self.retrievers[i].data) == list:
                        for continues_struct in self.retrievers[i].data:
                            length += continues_struct.get_length()
                    else:
                        length = self.retrievers[i].data.get_length()
                elif datatype == "str":
                    length = len(self.retrievers[i].data)
                total_length += length
        except TypeError:
            print(self.retrievers)
        return total_length

    def set_data_from_generator(self, generator):
        if self.parser:
            for i, retriever in enumerate(self.retrievers):
                try:
                    retriever.set_data(self.parser.retrieve_value(generator, retriever))
                except StopIteration as e:
                    print(f"[StopIteration] AoE2Piece.set_data_from_generator: \n\tRetriever: {retriever}\n")
                    raise StopIteration(e)

    def _entry_to_string(self, name, data, datatype):
        return "\t" + name + ": " + data + " (" + datatype + ")\n"

    def get_header_string(self):
        return "######################## " + self.piece_type + " ########################"

    def get_byte_structure_as_string(self, skip_retrievers=None):
        if skip_retrievers is None:
            skip_retrievers = []

        byte_structure = "\n" + self.get_header_string()

        for retriever in self.retrievers:
            if retriever.name in skip_retrievers:
                continue
            byte_structure += "\n"
            if type(retriever.data) is list and len(retriever.data) > 0:
                if isinstance(retriever.data[0], AoE2Piece):
                    for struct in retriever.data:
                        byte_structure += struct.get_byte_structure_as_string()
                    continue
            else:
                if isinstance(retriever.data, AoE2Piece):
                    byte_structure += retriever.data.get_byte_structure_as_string()
                    continue

            retriever_data_bytes = parser.retriever_to_bytes(retriever)
            if retriever_data_bytes is None:
                return byte_structure
            else:
                retriever_data_bytes = retriever_data_bytes.hex()

            retriever_short_string = retriever.get_short_str()

            retriever_data_hex = helper.create_textual_hex(retriever_data_bytes, space_distance=2,
                                                           enter_distance=24)

            if "\n" in retriever_data_hex:
                split_hex = retriever_data_hex.split("\n")
                if "\r\n" in retriever_short_string:
                    split_data_string = retriever_short_string.split("\r\n")

                    split_hex_length = len(split_hex)
                    split_data_string_length = len(split_data_string)
                    lines = max(split_hex_length, split_data_string_length)

                    combined_strings = []
                    for i in range(0, lines):
                        combined_strings.append(
                            helper.add_suffix_chars(split_hex[i] if i < split_hex_length else "", " ", 28) +
                            (split_data_string[i] if i < split_data_string_length else "")
                        )

                    byte_structure += "\n".join(combined_strings)
                else:
                    split_hex[0] = helper.add_suffix_chars(split_hex[0], " ", 28) + retriever_short_string
                    byte_structure += "\n".join(split_hex)
            else:
                byte_structure += helper.add_suffix_chars(retriever_data_hex, " ", 28) + retriever_short_string

        return byte_structure + "\n\n"

    def __str__(self):
        represent = self.piece_type + ": \n"

        for i, val in enumerate(self.retrievers):
            if type(self.retrievers[i].data) is list and len(self.retrievers[i].data) > 0:
                if isinstance(self.retrievers[i].data[0], AoE2Piece):
                    represent += "\t" + val.name + ": [\n"
                    for x in self.retrievers[i].data:
                        represent += "\t\t" + str(x)
                    represent += "\t]\n"
                else:
                    represent += self._entry_to_string(
                        val.name,
                        str(self.retrievers[i].data),
                        str(val.datatype.to_simple_string())
                    )
            else:
                if self.retrievers[i].data is not None:
                    data = self.retrievers[i].data
                else:
                    data = "<Empty>"
                represent += self._entry_to_string(val.name, str(data), str(val.datatype.to_simple_string()))

        return represent

    def __repr__(self):
        return type(self).__name__
